#include "world.hh"
#include <cmath>
#include <cstdlib>

World::World(int xsize, int ysize)
  : xsize(xsize), ysize(ysize), rock(xsize, ysize), water(xsize, ysize)
{
  const float ff = 100.0 / xsize;
  const float c = 0.05;
  const float af = 0.01;
  static const float o1 = 6.28 * (double(rand()) / RAND_MAX);
  static const float o2 = 6.28 * (double(rand()) / RAND_MAX);
  static const float o3 = 6.28 * (double(rand()) / RAND_MAX);
  for(int x = 1; x < xsize; x++)
    for(int y = 1; y < ysize; y++) {
      const float a = rock.height(x - 1, y);
      const float b = rock.height(x, y - 1);
      const float r = 2 * (double(rand()) / RAND_MAX) - 1; // ]-1,1[
      rock.height(x, y) = 
        1 * sin(ff * 0.04 * x + o1)
        + 2 * cos(ff * 0.09 * y + o2)
        + 1 * sin(ff * 0.05 * (x - y) + o3);
      //rock.height(x, y) += 0.3 * (af * a + (1-af) * b + (0.5 * fabs(a - b) + c) * r);
      rock.height(x, y) += c * r;
    }
}

vec3 World::normal(int x, int y) const {
  if(x == 0 || x == xsize - 1
     || y == 0 || y == ysize - 1)
    return vec3(0, 0, 1);
  const vec3 xv = get_point(x + 1, y) - get_point(x - 1, y);
  const vec3 yv = get_point(x, y + 1) - get_point(x, y - 1);
  return xv.cross_prod(yv).normalized();
}

vec2 World::gradient(const vec2 &p) const {
  if(p.x <= 0 || p.x >= xsize - 1
     || p.y <= 0 || p.y >= ysize - 1)
    return vec2(0, 0);

  const float dx = get_height(p.x + 1, p.y) - get_height(p.x - 1, p.y);
  const float dy = get_height(p.x, p.y + 1) - get_height(p.x, p.y - 1);
  return vec2(dx, dy);
}

void World::update() {
  const float d = 0.06;
  const int drops_per_frame = int(4 * xsize * ysize / 100.0 / 100.0);
  const int max_iterations = 100;
  const float radius = 3.0;

  for(int j = 0; j < drops_per_frame; j++) {
    HeightField tmp(xsize, ysize);
    //HeightField wtmp(xsize, ysize);
    vec2 p = vec2(xsize * (double(rand()) / RAND_MAX),
                  ysize * (double(rand()) / RAND_MAX));
    for(int i = 0; i < max_iterations; i++) {
      const vec2 gp = gradient(p);
      const float gp_mag = gp.length();
      const vec2 dp = gp.normalized();
    
      p = p - dp;
      if(p.x <= 0 || p.x >= xsize - 1
         || p.y <= 0 || p.y >= ysize - 1)
        break;;

      if(tmp.height(p.x, p.y) < -2 * d) {
        tmp.erode(p.x, p.y, radius, -d);
        //        wtmp.erode(p.x, p.y, radius, -2 * d);
        break;
      }

      tmp.erode(p.x, p.y, radius, d * gp_mag);
      //      wtmp.erode(p.x, p.y, radius, -d * gp_mag);
    }
  
    for(int x = 0; x < xsize; x++)
      for(int y = 0; y < ysize; y++) {
        rock.height(x, y) += tmp.height(x, y);
        //        water.height(x, y) *= 0.95;
        //water.height(x, y) += wtmp.height(x, y);
      }
  }

  for(int x = 0; x < xsize; x++)
    for(int y = 0; y < ysize; y++)
      water.height(x, y) = 0;

  for(int j = 0; j < 100 * drops_per_frame; j++) {
    vec2 p = vec2(xsize * (double(rand()) / RAND_MAX),
                  ysize * (double(rand()) / RAND_MAX));
    for(int i = 0; i < max_iterations; i++) {
      const vec2 gp = gradient(p);
      const vec2 dp = gp.normalized();
    
      p = p - dp;
      if(p.x <= 0 || p.x >= xsize - 1
         || p.y <= 0 || p.y >= ysize - 1)
        break;

      if(i == max_iterations - 1)
        water.erode(p.x, p.y, 3, -0.7 * d);
    }
  }
}
